(*
  4/19/98
  Database Desktop Lite 0.01
  calix@together.net
  Tested using Delphi 2 & Delphi 3 with BDE 4.5

  This is the UI for a simple Database Desktop application.  The application was developed to experiment with the BDE.
  My main intereste was it's ability to add, modify and delete fields in existing tables.

  Database Desktop Lite 0.01 has the following capabilities :
    Create a new table
    Pack a table
    Add, Modify and Delete fields from an existing table

  In the future, I hope to add the following capabilities :
    Add, Modify and Delete Primary Index of a table
    Generate source code that can perform the table modifications.
*)

unit uDBD;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, StdCtrls, Menus, Buttons, ExtCtrls, DB, DBTables,
  uCtDBD,
  BDE;

const
  GV_FIELDROSTERFLDNO = 0;
  GV_FIELDROSTERFLDNAME = 1;
  GV_FIELDROSTERTYPE = 2;
  GV_FIELDROSTERSIZE = 3;
  //GV_FIELDROSTERKEY = 4;

type
  TRestructForm = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    FieldRoster: TStringGrid;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Save1: TMenuItem;
    SaveAs1: TMenuItem;
    Open1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    New1: TMenuItem;
    Options1: TMenuItem;
    Pack1: TMenuItem;
    Help1: TMenuItem;
    About1: TMenuItem;
    ListBox1: TListBox;
    procedure FormShow(Sender: TObject);
    procedure FieldRosterKeyPress(Sender: TObject; var Key: Char);
    procedure FieldRosterKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Exit1Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure SaveAs1Click(Sender: TObject);
    procedure Save1Click(Sender: TObject);
    procedure New1Click(Sender: TObject);
    procedure Pack1Click(Sender: TObject);
    procedure FieldRosterSetEditText(Sender: TObject; ACol, ARow: Longint;
      const Value: string);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure FieldRosterDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure FieldRosterDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure FieldRosterSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
  public
    TableEditor: TCtTableEditor;
    // custom event handlers
    procedure RegisterRoster;
    procedure RegisterTable(Table: TTable);
    procedure Open(TblName: String);
    procedure Save(TblName: String);
  end;

var
  RestructForm: TRestructForm;

implementation

uses uAbout;

{$R *.DFM}

(*
  Register all the fields in the Roster
*)
procedure TRestructForm.RegisterRoster;
var
  i, j: Integer;
  //PrimaryFields: String;
begin
  with FieldRoster do begin
    //PrimaryFields := '';
    for i := 0 to RowCount-FixedRows-1 do begin
      for j := fldPDXCHAR to fldPDXUNICHAR do if (0 = CompareText(BDEFieldTypeName[PDX2BDEFieldType[j]], Cells[GV_FIELDROSTERTYPE, FixedRows+i])) then break;
      TableEditor.RegisterField(Cells[GV_FIELDROSTERFLDNAME, FixedRows+i], PDX2BDEFieldType[j], CtStrToInt(Cells[GV_FIELDROSTERSIZE, FixedRows+i]));
      (*
      if ('*' = Cells[GV_FIELDROSTERKEY, FixedRows+i]) then
        if ('' = PrimaryFields) then PrimaryFields := Cells[GV_FIELDROSTERFLDNAME, FixedRows+i]
        else PrimaryFields := PrimaryFields+';'+Cells[GV_FIELDROSTERFLDNAME, FixedRows+i];
      *)
    end;
    //if ('' <> PrimaryFields) then TableEditor.RegisterIndex('', PrimaryFields, [ixPrimary, ixUnique]);
  end;
end;

(*
  Display all the fields in the table in the roster
*)
procedure TRestructForm.RegisterTable(Table: TTable);
var
  i: Integer;
begin
  with FieldRoster do begin
    Row := FixedRows;
    Col := FixedCols;
    RowCount := FixedRows + Table.FieldCount;
    for i := 0 to Table.FieldCount-1 do begin
      with Table.Fields[i] do begin
        //if IsIndexField then Cells[GV_FIELDROSTERKEY, FixedRows+i] := '*' else Cells[GV_FIELDROSTERKEY, FixedRows+i] := '';
        Cells[GV_FIELDROSTERFLDNO, FixedRows+i] := IntToStr(i+1);
        Cells[GV_FIELDROSTERFLDNAME, FixedRows+i] := FieldName;
        Cells[GV_FIELDROSTERTYPE, FixedRows+i] := BDEFieldTypeName[DataType];
        Cells[GV_FIELDROSTERSIZE, FixedRows+i] := CtIntToStr(Size);
      end;
    end;
  end;
end;

(*
  Display the structure information of a table by opening the table,
  modifies the field roster to reflect table information
  and then closing the table.
*)
procedure TRestructForm.Open(TblName: String);
var
  Table: TTable;
begin
  if ('' = TblName) then exit;
  if FileExists(TblName) then begin
    Table := TTable.Create(nil);
    try
      with Table do begin
        Exclusive := TRUE;
        TableName := TblName;
        Active := TRUE;
      end;
      RegisterTable(Table);
    finally
      Table.Free;
    end;
  end;
end;

(*
  Using the structure information in the Table Roster create a new table.
*)
procedure TRestructForm.Save(TblName: String);
var
  Table: TTable;
begin
  if ('' = TblName) then exit;
  TableEditor.Clear;
  RegisterRoster;
  if FileExists(TblName) then begin
    Table := TTable.Create(nil);
    try
      with Table do begin
        Exclusive := TRUE;
        TableName := TblName;
        Active := TRUE;
        TableEditor.RegisterTable(Table);
        TableEditor.Execute;
      end;
    finally
      Table.Free;
    end;
  end else TableEditor.CreateTable(TblName);
end;

(*
  Initialize this form by setting the Table Roster headings,
  initializing the FieldTypeName array to contain all available field types
*)
procedure TRestructForm.FormShow(Sender: TObject);
var
  i: TFieldType;
begin
  with FieldRoster do begin
    RowCount := FixedRows + 1;
    Cells[GV_FIELDROSTERFLDNO, 0] := 'FNo';
    Cells[GV_FIELDROSTERFLDNAME, 0] := 'Field Name';
    Cells[GV_FIELDROSTERTYPE, 0] := 'Type';
    Cells[GV_FIELDROSTERSIZE, 0] := 'Size';
    //Cells[GV_FIELDROSTERKEY, 0] := 'Key';
  end;
  TableEditor.FIsPackTable := Pack1.Checked;
  for i := low(BDEFieldTypeName) to high(BDEFieldTypeName) do Listbox1.Items.Add(BDEFieldTypeName[i]);
  SetAutoSize(Self, TRUE);
end;

(*
  If the user presses the <ENTER> key in the Grid treat it as a tab with row-wrapping,
  inserting a new row if necessary.
*)
procedure TRestructForm.FieldRosterKeyPress(Sender: TObject; var Key: Char);
var
  ACol, ARow: Integer;
begin
  with FieldRoster do begin
    if Key = #13 then begin
      Key := #0;
      ARow := Row;
      ACol := (Col + 1) mod ColCount;
      if (ACol < FixedCols) then begin
        ACol := FixedCols;
        ARow := ARow + 1;
        if (ARow >= RowCount) then RowCount := ARow + 1;
      end;
      Col := ACol;
      Row := ARow;
    end;
  end;
end;

(*
  If the user presses the <CTRL+DEL> then delete the current row.
  If there is only one row remaining then clear the row, but don't delete it.
  If the user presses <CTRL+INS> then insert a new record.
  If the user presses <CTRL+D> then duplicate the above cell.
*)
procedure TRestructForm.FieldRosterKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  i, j: Integer;
  ARow, ACol: Integer;
begin
  with FieldRoster do begin
    if ([ssCtrl] = Shift) then begin
      if Key = VK_DELETE then begin
        // shift cells up
        for i := 0 to ColCount-1 do begin
          for j := Row to RowCount-2 do begin
            Cells[i, j] := Cells[i, j+1];
          end;
        end;
        if (RowCount > 1+FixedRows) then RowCount := RowCount - 1
        else for i := 0 to ColCount-1 do Cells[i, Row] := '';
        Col := FixedCols;
      end
      else if Key = VK_INSERT then begin
        // shift cells down
        RowCount := RowCount + 1;
        for i := 0 to ColCount-1 do begin
          for j := RowCount-1 downto Row do begin
            Cells[i, j] := Cells[i, j-1];
          end;
          Cells[i, Row] := '';
        end;
        Col := FixedCols;
      end else if UpCase(chr(Key)) = 'D' then begin
        if (Row > FixedRows) then begin
          Cells[Col, Row] := Cells[Col, Row-1];
          ARow := Row;
          ACol := (Col + 1) mod ColCount;
          if (ACol < FixedCols) then begin
            ACol := FixedCols;
            ARow := ARow + 1;
            if (ARow >= RowCount) then RowCount := ARow + 1;
          end;
          Col := ACol;
          Row := ARow;
        end;
      end;
    end;
  end;
end;

(*
  If the user selects Exit, then close this form.
*)
procedure TRestructForm.Exit1Click(Sender: TObject);
begin
  Close;
end;

(*
  If the user selects Open, then execute the Open Dialog.
  If a file is selected, then execute the open command.
*)
procedure TRestructForm.Open1Click(Sender: TObject);
begin
  if OpenDialog1.Execute then begin
    try
      Open(OpenDialog1.Filename);
      SaveDialog1.Filename := OpenDialog1.FileName;
      Caption := OpenDialog1.FileName;
    finally
    end;
  end;
end;

(*
  If the user selects SaveAs, then execute the Save Dialog.
  If a file name is specified, then save the table.
*)
procedure TRestructForm.SaveAs1Click(Sender: TObject);
begin
  if SaveDialog1.Execute then begin
    OpenDialog1.Filename := SaveDialog1.FileName;
    Caption := SaveDialog1.FileName;
  end;
  Save(SaveDialog1.Filename);
end;

(*
  If the user selects Save then save the table.
  If the SaveDialog filename has not been specified then call SaveAs.
*)
procedure TRestructForm.Save1Click(Sender: TObject);
begin
  if ('' <> SaveDialog1.Filename) then Save(SaveDialog1.Filename)
  else SaveAs1Click(Sender);
end;

(*
  If the user selects New, then clear the Field Roaster.
*)
procedure TRestructForm.New1Click(Sender: TObject);
var
  i: Integer;
begin
  with FieldRoster do begin
    RowCount := FixedRows + 1;
    for i := FixedCols to ColCount do Cells[i, FixedRows] := '';
  end;
end;

procedure TRestructForm.Pack1Click(Sender: TObject);
begin
  Pack1.Checked := NOT(Pack1.Checked);
  TableEditor.FIsPackTable := Pack1.Checked;
end;

procedure TRestructForm.FieldRosterSetEditText(Sender: TObject; ACol,
  ARow: Longint; const Value: string);
var
  i: Integer;
begin
  if (ARow < FieldRoster.FixedRows) then exit;
  case ACol of
    GV_FIELDROSTERTYPE:
    begin
      i := ctGetIndex(FieldRoster.Cells[GV_FIELDROSTERTYPE, ARow], BDEFieldTypeName);
      if (i >= 0) AND NOT BDEFieldTypeSize[TFieldType(i)] then FieldRoster.Cells[GV_FIELDROSTERSIZE, ARow] := '';
    end;
    GV_FIELDROSTERSIZE:
    begin
      i := ctGetIndex(FieldRoster.Cells[GV_FIELDROSTERTYPE, ARow], BDEFieldTypeName);
      if (i >= 0) AND NOT BDEFieldTypeSize[TFieldType(i)] then FieldRoster.Cells[GV_FIELDROSTERSIZE, ARow] := '';
    end;
  end;
end;


procedure TRestructForm.FormCreate(Sender: TObject);
begin
  TableEditor := TCtTableEditor.Create;
end;

procedure TRestructForm.FormDestroy(Sender: TObject);
begin
  TableEditor.Free;
end;

procedure TRestructForm.About1Click(Sender: TObject);
begin
  AboutBox := TAboutBox.Create(Application);
  try
    AboutBox.ShowModal;
  finally
    AboutBox.Free;
  end;
end;

procedure TRestructForm.FieldRosterDragDrop(Sender, Source: TObject; X,
  Y: Integer);
var
  ACol, ARow: Longint;
begin
  with FieldRoster do begin
    MouseToCell(X, Y, ACol, ARow);
    if (Source is TListBox) then begin
      if (0 > ARow) OR (ARow >= RowCount) then begin
        ARow := RowCount;
        RowCount := RowCount + 1;
      end;
      with (Source as TListBox) do Cells[GV_FIELDROSTERTYPE, ARow] := Items[ItemIndex];
      Col := GV_FIELDROSTERTYPE;
      Row := ARow;
    end;
  end;
end;

procedure TRestructForm.FieldRosterDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
var
  ACol, ARow: Longint;
begin
  with FieldRoster do begin
    MouseToCell(X, Y, ACol, ARow);
    Accept := (Source is TListBox);
  end;
end;

procedure TRestructForm.FieldRosterSelectCell(Sender: TObject; Col,
  Row: Longint; var CanSelect: Boolean);
var
  i: Integer;
begin
  case Col of
    GV_FIELDROSTERSIZE:
    begin
      i := ctGetIndex(FieldRoster.Cells[GV_FIELDROSTERTYPE, Row], BDEFieldTypeName);
      CanSelect := (i < 0) OR BDEFieldTypeSize[TFieldType(i)];
    end;
  end;
end;

end.
